--[[ <HCExtension>
@name				Content Master
@author			nadenj@mail.ru
@version			0.48
@description		Content Editing Engine
@min_HC_version	1.0.0.263
@event				Init/init
@event				RequestHeaderReceived/RequestHeaderReceived
@event				Options/options
@event				Destroy/destroy
</HCExtension> ]]


function to_log(...)
	if not _CM_.LogOn then return end
	-- hc.put_to_log(...)				--    HC
	CM_log_fragment = (CM_log_fragment or '') .. table.concat(arg) .. '\r\n'	--     Content Master.log
end

function toregexp(str, hc)	--      
	str = str:gsub('[.+|)(}{%][\\$^]', '\\%0')	--  
	str = str:gsub('%?', hc and '.' or '\\?')
	str = str:gsub('^%*', '')	--   '*'
	str = str:gsub('%*$', '')	--   '*'
	return (str:gsub('%*', '.*?'))
end

function regexp_from_list(lst)	--      

	local function delist(lst)
		if #lst<2 then return lst[1] end
		local lst_tmp = {}
		for _,v in ipairs(lst) do
			local index = re.find(v, nil, 1)
			lst_tmp[index] = lst_tmp[index] or {}
			table.insert(lst_tmp[index], re.substr(0))
		end
		local s = {}
		for ind,v in pairs(lst_tmp) do
			table.insert(s, ind~='' and (ind .. delist(v)) or (#v>1 and ('(?:' .. table.concat(v, '|') .. ')') or v[1]))
		end
		return #s>1 and ('(?:' .. table.concat(s, '|') .. ')') or s[1]
	end

	if #lst==0 then return '' end
	re.set_regex([=[\A((?>[^\\*+{}?[\]()|]|\\[.?wd()])[?*+]*+(?![{}])|)\K(?:(?(DEFINE)(?<qe>\\Q.*?\\E)(?<square_brackets>\[(?>\\.|[^\]])*+\])(?<round_brackets>\((?>(?&qe)|(?&square_brackets)|(?&round_brackets)|\\.|[^)])*+\)))(?>(?&qe)|(?&round_brackets)|(?&square_brackets)|\\.|[^|])*+|\A.*)\z]=])
	return string.gsub(delist(lst), '^%(%?:(.*)%)$', '%1')
end

function SaveParams()
	if not _CM_ then
		_CM_ = hc.get_global('CM_DATA')
	end
	local f, err = io.open(hc.script_name:match('.*%.') .. 'ini', 'wb') --     Content_Master.ini
	if not f then
		hc.put_to_log('Content Master:\r\n', err)
		hc.put_msg('Content Master:\r\n', err)
		return
	end
	local H = hc.get_global('CM_Hits')
	for i,v in ipairs(_CM_.Data) do
		f:write(v.Name .. '=' .. tostring(v.On) .. '\r\n')
		if v.File then
			f:write(v.Name:match('%s+(.+)') .. ' files=' .. v.File .. '\r\n')
		end
		for j,rule in ipairs(v.Rules) do
			f:write(rule.Name .. '=' .. tostring(rule.On) .. ',' .. (H[i][j] or 0) .. '\r\n')
		end
	end
	f:write(' =' .. tostring(_CM_.LogOn))
	f:write('\r\n  =' .. _CM_.CM_Off_Hotkey)
	f:write('\r\n[GLOBAL EXCEPTIONS]\r\n' .. table.concat(_CM_.User_Exceptions_List, '\r\n') .. '\r\n')
	f:close()
end

local function process_update(lst)
	hc.set_global('CM_UpdateList', lst)
	if not re.find(lst, [[^(?!(?>[^\r\n]*?\#~\#){3})]]) then
		local s = re.replace(lst, [[^((?>[^\r\n]*?\#~\#)){3}\s*?(^|\z)|^(?>[^\r\n]*?(?=\#~\#))\K...(?1){2}1$]], '', true)
		if s:find('%S') then
			hc.reload_extension('Content Master')
			hc.put_msg(20, '--- Content Master ---\r\n:\r\n', s)
		else
			hc.put_msg(5, '--- Content Master ---\r\n  ')
		end
	end
end

function callout3(version, callout_number, offset_vector, start_match,
	current_position, capture_top, capture_last, pattern_position, next_item_length)
	return 1
end

CM_UpdatePostfix = ' HTTP/1.1\r\nCMUpdate:'
CM_UpdateCfgURL = 'http://nadenj.narod.ru/ContentMaster_Source/cm_update.cfg'


function init()
	dofile(hc.script_name:match('.*\\') .. 'cm_init.lua')
end


function RequestHeaderReceived()
	CMUPDATE = hc.request_header:find(CM_UpdatePostfix, 1, true)
	if CMUPDATE then	--     
		if hc.url==CM_UpdateCfgURL then
			hc.call_me_for('BeforeAnswerBodySend', 'BeforeAnswerBodySend')	--    
		else
			hc.request_header = hc.request_header:gsub('\r\n$', 'If-Modified-Since: ' ..
				hc.systime_to_str(hc.file_last_modified_time(hc.script_name:match('.*\\') ..
				re.find(hc.get_global('CM_UpdateList'), [[^(?>(.*?)#~#){2}]] .. toregexp(hc.url), 1)) or 0) .. '\r\n\r\n', 1)
			hc.request_header = re.replace(hc.request_header, [[^Host:\s*+\K.*?$]], hc.url:match('^https?://(.-)[/%?]'))
			hc.call_me_for('BeforeAnswerHeaderSend', 'BeforeAnswerHeaderSend')	--    
		end
		return
	end
	--           ,  () 
	if not _CM_Connection_Id or (_CM_Connection_Id < hc.get_global('CM_GLOBAL_ID')) then
		_CM_ = hc.get_global('CM_DATA')
		_CM_Connection_Id = hc.get_global('CM_GLOBAL_ID')
		_CM_URLCache = {}
	end
	if	_CM_.CM_Off_Hotkey~='' then
		local kbd_now = hc.get_keyboard_state()
		local block = true
		for key in _CM_.CM_Off_Hotkey:gmatch('%S+') do
			if not re.find(kbd_now, '\\b' .. key .. '\\b') then
				block = false
				break
			end
		end
		if block then return end	--     ,  
	end
	if	re.find(hc.url, _CM_.URL_Exception) then return end	--      ,  
	--    ,  WebWarper  OperaTurbo
	_CM_.URL = re.replace(hc.url, [[^http://\K\w++-turbo[^.]*+\.opera-mini\.net:80/]])
	_CM_.URL = re.replace(_CM_.URL, [[^http://\Kwww\.webwarper\.net/ww/(?:~clientscriptgz/)?(?=(?:[^/])+\.(?:\w)+/)]])

	local url_for_HC = re.replace(_CM_.URL, [[^https?://(www\.)?]])
	
	-- local CacheItem = hc.get_global_table_item('_CM_URLCache', _CM_.URL)
	-- if not CacheItem then 
	if not _CM_URLCache[_CM_.URL] then 
		re.set_callout('callout3')
		for i,v in ipairs(_CM_.Data) do
			if v.On then
				local URL = (v.Type=='HandyCache') and url_for_HC or _CM_.URL
				for j,rule in ipairs(v.Rules) do
					if rule.Method and rule.On and rule.Method:find('block', 1, true) then
						local t, PCRE_ErrMsg = re.find(URL, rule.SubFind)
						if t then
							re.set_callout(nil)
							local H = hc.get_global('CM_Hits')
							H[i][j] = H[i][j] + 1
							hc.set_global('CM_Hits', H)
							hc.action = 'stop'
							hc.monitor_string = hc.monitor_string .. 'CM:' .. v.Type .. ' block'
							_CM_URLCache[_CM_.URL] = { Blocked=true, BlockType=v.Type }
							-- hc.set_global_table_item('_CM_URLCache', _CM_.URL, { Blocked=true, BlockType=v.Type })
							return
						elseif not PCRE_ErrMsg:find('PCRE_ERROR_NOMATCH', 1, true) then
							hc.put_msg('Content Master: ', v.Name, '\r\n', rule.Name, ': ', PCRE_ErrMsg)
							hc.put_to_log('*1* ', 'Content Master: ', v.Name, '\r\n', rule.Name, ': ', PCRE_ErrMsg,
								'\r\nURL = ', URL, '\r\n', rule.SubFind)
						end
					end
				end
			end
		end
		re.set_callout(nil)
		_CM_URLCache[_CM_.URL] = { Blocked=false }
		-- hc.set_global_table_item('_CM_URLCache', _CM_.URL, { Blocked=false })
	elseif _CM_URLCache[_CM_.URL].Blocked then
	-- elseif CacheItem.Blocked then
		local H = hc.get_global('CM_Hits')
		H[i][j] = H[i][j] + 1
		hc.set_global('CM_Hits', H)
		hc.action = 'stop'
		hc.monitor_string = hc.monitor_string .. 'CM:' .. _CM_URLCache[_CM_.URL].BlockType .. ' block'
		-- hc.monitor_string = hc.monitor_string .. 'CM:' .. CacheItem.BlockType .. ' block'
		return
	end
	hc.call_me_for('BeforeAnswerHeaderSend', 'BeforeAnswerHeaderSend')	--    
end

-- *********************************************************************************************

function BeforeAnswerHeaderSend()
	if CMUPDATE then
		if hc.answer_header:match('%s(%d+)')=='200' then
			hc.call_me_for('BeforeAnswerBodySend', 'BeforeAnswerBodySend')		--    
		else
			process_update(re.replace(hc.get_global('CM_UpdateList'), [[\#~\#]] .. toregexp(hc.url) .. '\\K', '#~#'))
		end
		return
	end
	--    200    html,    
	if not re.find(hc.answer_header, [[\A\S++\s++200\D.*?^Content-Type:\s*+text/html]]) then return end
	_CM_.Host = re.find(_CM_.URL, [[^\w+://[^:/?]*]]) .. '/'
	local domain = re.find(_CM_.Host, [[^\w+://\K.*(?=/$)]])
	for _,v in ipairs(_CM_.Data) do
		if v.On then
			for _,rule in ipairs(v.Rules) do
				if rule.On then
					if rule.__AdBlockPlus_CSS_With_Domain_List then
						local doms, n = re.replace(toregexp(domain), [[.*?\.]], [[\0)?]], true)
						local Rest_List = re.replace(rule.__AdBlockPlus_CSS_With_Domain_List,
							[[\G\s*+(*COMMIT)]] .. string.rep('(?:', n+1) .. doms ..
							[[(?:,|(?=#)).*?\#\#(.*?\r\n)|[^,#]++(?:,|\#.*?$))]], '\\1', true)
						local tmp = ' AdBlockPlus CSS SELECTORS FOR ' .. domain
						rule.ReplaceByDomain = Rest_List:find('%S') and	('<!--' .. tmp ..
							' (INSERTED BY CONTENT MASTER): -->\r\n<style type="text/css">\r\n' .. Rest_List:match('(.-)%s*$') ..
							'\r\n</style>\r\n<!-- END' .. tmp .. ' -->\r\n\r\n') or nil
					elseif rule.__CMReplacement then	--       
						function _CM_Replacement_Callout(version, callout_number, offset_vector, start_match,
							current_position, capture_top, capture_last,	pattern_position, next_item_length)
							if callout_number==252 then
								re.select_engine(3)
								local tmp = re.find(_CM_.URL, rule.__CMReplacement:sub(offset_vector[2*capture_last]+1,
									offset_vector[2*capture_last+1]))
								re.select_engine(1)
								if not tmp then return 1 end
							end
							return 0
						end
						re.set_callout('_CM_Replacement_Callout')
						local regexp = re.replace(rule.__CMReplacement, [[^(?>(.*?)#~#)(?:(?C252)|.*?(?:^|\z))]], '', true)
						re.set_callout(nil)
						rule.Find = '(?|' .. regexp:match('(.-)%s*$'):gsub('[\r\n]+', '|') .. ')'
					end
				end
			end
		end
	end
	hc.call_me_for('BeforeAnswerBodySend', 'BeforeAnswerBodySend')		--    
end

-- *********************************************************************************************

function BeforeAnswerBodySend()
-- -------------------------  --------------------------------
	if CMUPDATE then
		CMBODY = (CMBODY or '') .. hc.answer_body
		if hc.last_part then
			if hc.url==CM_UpdateCfgURL then
				CMBODY = re.replace(CMBODY, [[^\s*+|$\s*+\z|\s*?$]], '', true)
				hc.set_global('CM_UpdateList', CMBODY)
				for url in CMBODY:gmatch('#~#.-#~#(%S+)') do
					hc.execute_cmd('LoadURL ' .. url .. CM_UpdatePostfix)
				end
			elseif CMUPDATE then
				local lst = re.replace(hc.get_global('CM_UpdateList'), [[\#~\#]] .. toregexp(hc.url) .. '\\K', '#~#1')
				local fname = hc.script_name:match('.*\\') ..
					re.find(hc.get_global('CM_UpdateList'), [[^(?>(.*?)\#~\#){2}]] .. toregexp(hc.url), 1)
				local f = io.open(fname, 'wb')
				if f then
					f:write(CMBODY)
					f:close()
				end
				process_update(lst)
			-- else
				-- local f, err = io.open(hc.script_name, 'rb')
				-- local code
				-- if f then
					-- code = f:read('*a')
					-- f:close()
				-- else
					-- hc.put_msg('Content Master. :\r\n', err)
					-- return
				-- end
				-- local Local_CM_Version = re.find(code, [[@version\s*+\K\S++]]) or 0
				-- local Server_CM_Version = re.find(CMBODY, [[@version\s*+\K\S++]]) or 0
				-- local Min_HC_Version = re.find(CMBODY, [[@min_HC_version\s*+\K\S++]]) or '0.0.0.0'
				-- local HC_Version = hc.version_number
				-- if Local_CM_Version<Server_CM_Version and HC_Version>=Min_HC_Version then
					-- f = io.open(hc.script_name, 'wb')
					-- if f then
						-- f:write(CMBODY)
						-- f:close()
						-- hc.reload_extension('Content Master')
						-- hc.put_msg('Content Master    ', Server_CM_Version)
					-- else
						-- hc.put_msg('Content Master:\r\n   .     ')
					-- end
				-- elseif Local_CM_Version>=Server_CM_Version then
					-- hc.put_msg('Content Master\r\n ', Local_CM_Version, '\r\n  ')
				-- else
					-- hc.put_msg('Content Master\r\n HandyCache  ', HC_Version,
						-- '\r\n      ', Min_HC_Version)
				-- end
			end
			CMBODY = nil
		end
		return
	end

-- -------------------------    --------------------------------
	-- to_log('\r\n  ...')
	local ToNextItem = { hc.answer_body }
	CM = CM or {Freezed={}, ReplCount=0}
 -- hc.put_to_log('\r\n 80     ', #ToNextItem[1], ':\r\n', ToNextItem[1]:sub(1,80))
 -- hc.put_to_log('\r\n 150   :\r\n', ToNextItem[1]:sub(-150))
	local replaces_in_chunk = 0	--     
	local partial = not hc.last_part and 'hard'

	_CM_Rule_Functions = {}
	function _CM_Rule_Functions.HC_SiteColumn (regexp)
		local URL = (hc.event=='RequestHeaderReceived') and re.find(hc.request_header, [[^Referer:\s*+\K.*?$|\z]]) or hc.url
		return (re.find(URL, regexp) and 0 or 1)
	end
	function _CM_Rule_Functions.TAG_NAME (tag)
		if tag=='third-party' then
			return hc.url:match('^https?://[^/?:]*')~=_CM_Tested_URL:match('^https?://[^/?:]*') and 0 or 1
		elseif tag=='~third-party' then
			return hc.url:match('^https?://[^/?:]*')~=_CM_Tested_URL:match('^https?://[^/?:]*') and 1 or 0
		end
		return re.find(_CM_Previous, tag) and 0 or 1
	end

	re.select_engine(1)
	re.set_callout('callout1')
	for ord_i,CurrentOrder in ipairs(_CM_.Orders) do	--     
		if partial and not ToNextItem[1] then break end
		for i,v in ipairs(_CM_.Data) do			--    
			if partial and not ToNextItem[1] then break end
			if v.On then								--    ,   
				if not CM.Freezed[i] then
					CM.Freezed[i] = {}
				end
				for j,rule in ipairs(v.Rules) do		--   
					if partial and not ToNextItem[1] then break end
					--   ()  ,   
					if rule.On and not rule.__Off and CurrentOrder==rule.Order and rule.Method:find('edit', 1, true) then
						local log_msg = ''	--        
						ToNextItem = { table.concat(ToNextItem) }
						local Subj = CM.Freezed[i][j] and (CM.Freezed[i][j] .. ToNextItem[1]) or ToNextItem[1]	--  
						local Replace_Str
						if rule.Replace then
							Replace_Str = rule.ReplaceByDomain and (rule.Replace .. rule.ReplaceByDomain) or rule.Replace
						end

function callout1(version, callout_number, offset_vector, start_match, current_position,
	capture_top, capture_last, pattern_position, next_item_length)

	if callout_number == 254 then
 		local URL
		local ind = 2*capture_last
		_CM_Tested_URL = Subj:sub(offset_vector[ind]+1, offset_vector[ind+1])	--  
		re.select_engine(2)
		if rule.Type=='URL' then
			local u,n = _CM_Tested_URL:gsub('%.%./', '')
			if n>0 then 	--     '../../folder/file.jpg'
				_CM_Tested_URL = (re.find(_CM_.URL, [[\A[^/:]++://(?:[^?/]+/)+?(?=(?:[^?/]++/){]] .. n ..
					[[}[^?/]*+(?:\?|\z))]]) or _CM_.Host) .. u
			else	--  url ,    
				_CM_Tested_URL = re.replace(_CM_Tested_URL, [[\A(?>\.?/|(?!\w++:))]], _CM_.Host)
			end
			--    ,  WebWarper'
			_CM_Tested_URL = re.replace(_CM_Tested_URL,
				[[\Ahttp://\Kwww\.webwarper\.net/ww/(?:~clientscriptgz/)?(?=[^/]+\.\w+/)]])
			URL = _CM_Tested_URL
			if v.Type=='HandyCache' then
				_CM_Tested_URL = re.replace(_CM_Tested_URL, [[\Ahttps?://(?:www\.)?]])	--     URL 
			end
		end
		_CM_Previous = (capture_last>1) and Subj:sub(offset_vector[ind-2]+1, offset_vector[ind-1]) or ''
		-- local CacheItem = hc.get_global_table_item('_CM_URLCache', URL)
		-- if URL and CacheItem then
		if URL and _CM_URLCache[URL] then
if URL then hc.put_to_log(rule.Name, ':  ', URL, '  (CACHED)') end
			re.select_engine(1)
			if _CM_URLCache[URL].Blocked then
			-- if CacheItem.Blocked then
				if _CM_Previous:match('^style') then	--   <style>    /*...*/
					Replace_Str = Replace_Str:gsub('<!%-%-', '/*')
					Replace_Str = Replace_Str:gsub('%-%->', '*/')
				end
				return 0
			else
				return 1
			end
		else
if URL then hc.put_to_log(rule.Name, ':  ', URL) end
			re.set_callout('callout2')
			local t, PCRE_ErrMsg = re.find(_CM_Tested_URL, rule.SubFind)	--  
			re.set_callout(nil)
			if t then
				if _CM_Previous:match('^style') then	--   <style>    /*...*/
					Replace_Str = Replace_Str:gsub('<!%-%-', '/*')
					Replace_Str = Replace_Str:gsub('%-%->', '*/')
				end
				if URL then
					_CM_URLCache[URL] = { Blocked=true, BlockType=v.Type }
					-- hc.set_global_table_item('_CM_URLCache', URL, { Blocked=true, BlockType=v.Type })
				end
				if rule.Name=='      ()' then
					Replace_Str = Replace_Str:gsub('keyword', 'keyword: ' .. re.substr(0))
				end
				re.select_engine(1)
				return 0
			elseif PCRE_ErrMsg:find('PCRE_ERROR_NOMATCH', 1, true) then
				if URL and (rule.Name==_CM_.LastURLRule) then
					_CM_URLCache[URL] = { Blocked=false }
					-- hc.set_global_table_item('_CM_URLCache', URL, { Blocked=false })
				end
				re.select_engine(1)
				return 1
			else
				log_msg = _CM_.LogOn and (log_msg .. '\r\n   : ' .. PCRE_ErrMsg .. '\r\n\r\n:\r\n' .. rule.SubFind)
				hc.put_msg('Content Master: ', v.Name, '\r\n', rule.Name, ': ', PCRE_ErrMsg)
				hc.put_to_log('*5* ', 'Content Master: ', v.Name, '\r\n', rule.Name, ': ', PCRE_ErrMsg)
				re.select_engine(1)
				return -1
			end
		end
	elseif callout_number == 253 then
		Replace_Str = string.gsub(rule.Find:match('[^%)]*', pattern_position+3), '####', ')')
		Replace_Str = Replace_Str:gsub('\\1', '\\' .. capture_last)
	elseif callout_number == 251 then
		Replace_Str = string.gsub(rule.Find:match('[^%)]*', pattern_position+3), '####', ')')
	end
	return 0
end

function callout2(version, callout_number, offset_vector, start_match,
	current_position, capture_top, capture_last, pattern_position, next_item_length)

	if callout_number == 251 then
		if rule.Functor and _CM_Rule_Functions[rule.Functor] then
			re.select_engine(3)
			local ret = _CM_Rule_Functions[rule.Functor](string.gsub(rule.SubFind:match('[^%)]+', pattern_position+3), '####', ')'))
			re.select_engine(2)
			return ret
		end
	end
	return 0
end

						local start_pos = 1		--      
						ToNextItem = {}	--         
						local hit_count = 0
						local Replace_Str_Store = Replace_Str
						local t, PCRE_ErrMsg = re.match(Subj, rule.Find, start_pos, partial)
						local function decapture(s)
							local n = tonumber(s:sub(2,2))
							return Subj:sub(t[n].begin_pos, t[n].end_pos)
						end
						while t do
							local repl = Replace_Str:gsub('\\%d', decapture)
							table.insert(ToNextItem, Subj:sub(start_pos, t[0].begin_pos-1))
							table.insert(ToNextItem, repl)
							hit_count = hit_count + 1
							log_msg = _CM_.LogOn and (rule.Counted and (log_msg .. '\r\n\r\n:\r\n' ..
								Subj:sub(t[0].begin_pos, t[0].end_pos) .. '\r\n :\r\n' .. repl) or log_msg)
							start_pos = t[0].end_pos + 1
							if rule.Once then
								rule.__Off = true
								break
							end
							Replace_Str = Replace_Str_Store
							t, PCRE_ErrMsg = re.match(nil, nil, start_pos, partial)
						end	-- while t do
						if rule.Counted and hit_count>0 then
							replaces_in_chunk = replaces_in_chunk + hit_count
							local H = hc.get_global('CM_Hits')
							H[i][j] = H[i][j] + hit_count
							hc.set_global('CM_Hits', H)
						end
						if PCRE_ErrMsg and PCRE_ErrMsg:find('PCRE_ERROR_PARTIAL', 1, true) then
							local last_bumpalong = re.offsets_vector(0)
							table.insert(ToNextItem, Subj:sub(start_pos, last_bumpalong-1))
							CM.Freezed[i][j] = Subj:sub(last_bumpalong)
						else
							table.insert(ToNextItem, Subj:sub(start_pos))
							CM.Freezed[i][j] = nil
							if PCRE_ErrMsg and not PCRE_ErrMsg:find('PCRE_ERROR_NOMATCH', 1, true) then
								log_msg = _CM_.LogOn and (log_msg .. '\r\n   Find: ' .. PCRE_ErrMsg .. '\r\n' ..
									rule.Find	.. '\r\n\r\n  :' .. Subj:sub(start_pos))
								hc.put_msg('Content Master: ', v.Name or '', '\r\n', rule.Name or rule.Find, ': ', PCRE_ErrMsg)
								hc.put_to_log('*2* ', 'Content Master: ', v.Name or '', '\r\n', rule.Name, '\r\n', rule.Find, '\r\n', PCRE_ErrMsg)
							end
						end
						if _CM_.LogOn and log_msg~='' then
							to_log('\r\n', string.rep('-', 60), '\r\n: ', v.Name or '', ' / ', rule.Name or rule.Find, log_msg)
						end
					end	--   ()  
					rule.__Off = rule.__Off and not (hc.last_part and ord_i==#_CM_.Orders)
				end -- for j,rule
			end	--     
		end	-- for i,v
	end	-- for ord_i,CurrentOrder
	re.set_callout(nil)

	CM.ReplCount = CM.ReplCount + replaces_in_chunk
	if hc.last_part then
		hc.monitor_string = hc.monitor_string .. 'CM:' .. CM.ReplCount .. ' hits'
		if _CM_.LogOn and CM_log_fragment then		--   ,     
			local f = io.open(hc.script_name:match('.*%.') .. 'log', 'ab')
			if not f then
				hc.put_to_log('Content Master:\r\n', err)
				hc.put_msg('Content Master:\r\n', err)
				return
			end
			f:write('=====  ', os.date(), '    URL  ', hc.url, '  =====\r\n',
				CM_log_fragment,
				'\r\n=====   URL   =====\r\n', string.rep('*', 80), '\r\n\r\n')
			f:close()
			CM_log_fragment = nil
		end
		CM = nil
	end
	hc.answer_body = table.concat(ToNextItem)
	-- to_log('\r\n   ')
end


-- -------------------------   --------------------------------
function options()
	dofile(hc.script_name:match('.*\\') .. 'cm_options.lua')
end


function destroy()
	if not _CM_ then
		_CM_ = hc.get_global('CM_DATA')
	end
	SaveParams()	--    
	--    
	hc.set_global('CM_Hits')
	hc.set_global('CM_DATA')
	hc.set_global('CM_GLOBAL_ID')
--	hc.set_global_table_item('_CM_URLCache')
end
